open Parser
open Mlsyntax
open Lexer

exception No_match
  
exception Type_error

exception Implementation_error

(**********)
(* VALUES *)
(**********)
(* defines the possible values of evaluating an expression *)
type evalvalue = 
  | Bool of bool
  | Int of Big_int.big_int
  | Fun of string * expr * evalenv (* closure *)
  | Fix of string * expr * evalenv
  | Location of int
  | Unit

(* expressions are evaluated under such an environment *)
and evalenv = (string * evalvalue) list

(* print the result of the evaluation *)
let print_value value = match value with
  | Bool(b) -> print_string (string_of_bool b)
  | Int(n) -> print_string (Big_int.string_of_big_int n)
  | Fun(_, _, _) -> print_string "<fun>"
  | Fix(_, _, _) -> print_string "<funrec>"
  | Location(_) -> print_string "<loc>"
  | Unit -> print_string "()"

(* print an environment *)
let rec print_env (env : evalenv) = match env with
  | ((name, value) :: rest) -> (
      print_string (name ^ " = ");
      print_value value;
      print_newline ()
    )
  | [] -> ()

(* add a binding to an environment *)
let push (env : evalenv) name value = (name, value) :: env

(* search the environment for a binding *)
let rec lookup name (env : evalenv) = match env with
  | (name', value) :: rest -> (
      if (name = name') then
	value
      else
	lookup name rest
    )
  | [] -> raise Not_found


(**********)
(* MEMORY *)
(**********)
(* the "store" or the set of allocated memory locations *)
(* there is (currently) no garbage collection *)
let evalmem = Hashtbl.create 16

(* how many memory locations have been created *)
let counter = ref 0

(* generate fresh names for memory locations *)
let new_loc () = (counter := !counter + 1; !counter)

(* add a value to a fresh memory location *)
let push_mem (value : evalvalue) =
  let loc = new_loc () in
    (Hashtbl.add evalmem loc value; loc)

(* search the memory for a certain location *)
let search_mem loc = Hashtbl.find evalmem loc

(* update a memory location *)
let rec update_mem loc value = Hashtbl.replace evalmem loc value


(********************)
(* HELPER FUNCTIONS *)
(********************)
let f_and x y = if x then y else false

let f_or x y = x or y

let f_xor x y = x != y 

let f_less = Big_int.lt_big_int

let f_greater = Big_int.gt_big_int

let f_leq = Big_int.le_big_int

let f_geq = Big_int.ge_big_int

(************************)
(* EVALUATION FUNCTIONS *)
(************************)

(* evaluate the expression consisting of applying f to the evaluation of e1 and e2 *)
let rec eval_binary_int e1 e2 env f =
  match eval e1 env with
    | Int(n1) -> (
        match eval e2 env with
          | Int(n2) -> Int(f n1 n2)
          | _ -> raise Type_error
      )
    | _ -> raise Type_error

and eval_binary_intbool e1 e2 env f =
  match eval e1 env with
    | Int(n1)  -> (
        match eval e2 env with
          | Int(n2) -> Bool(f n1 n2)
          | _ -> raise Type_error
      )
    | _ -> raise Type_error

and eval_binary_bool e1 e2 env f =
  match eval e1 env with
    | Bool(b1)  -> (
        match eval e2 env with
          | Bool(b2) -> Bool(f b1 b2)
          | _ -> raise Type_error
      )
    | _ -> raise Type_error

and print_expr e =
  match e with
    | (_, VAR(s)) -> print_string s
    | (_, FUN(s, e')) -> (print_string ("fun " ^ s ^ " -> "); print_expr e')
    | (_, APP(e1, e2)) -> (print_expr e1; print_string " ("; print_expr e2; print_string ")")
    | (_, ITE(e0, e1, e2)) -> (
	print_string "if "; print_expr e0; print_string " then ";
	print_expr e1; print_string " else "; print_expr e2)
    | (_, EQUAL(e0, e1)) -> (
	print_expr e0; print_string " = "; print_expr e1)
    | (_, PLUS(e0, e1)) -> (
	print_expr e0; print_string " + "; print_expr e1)
    | (_, MINUS(e0, e1)) -> (
	print_expr e0; print_string " - "; print_expr e1)
    | (_, MULT(e0, e1)) -> (
	print_expr e0; print_string " * "; print_expr e1)
    | (_, LESS(e0, e1)) -> (
	print_expr e0; print_string " < "; print_expr e1)
    | (_, GREATER(e0, e1)) -> (
	print_expr e0; print_string " > "; print_expr e1)
    | (_, UMINUS(e0)) -> (
	print_string " - "; print_expr e0)
    | (_, OR(e0, e1)) -> ( print_string "("; print_expr e0; print_string ")";
			   print_string " or ";
			   print_string "("; print_expr e1; print_string ")")
    | (_, LETREC(s, e0, e1)) -> (
	print_string ("letrec " ^ s ^ " = "); print_expr e0;
	print_string " in "; print_expr e1
      )
    | (_, LET(s, e0, e1)) -> (
	print_string ("let " ^ s ^ " = "); print_expr e0;
	print_string " in "; print_expr e1
      )
    | (_, REF(e0)) -> (
	print_string "ref "; print_expr e0
      )
    | (_, BANG(e0)) -> (
	print_string "!("; print_expr e0;
	print_string ")"
      )
    | (_, ASSIGN(e0, e1)) -> (
	print_string "("; print_expr e0; print_string ") := ";
	print_string "("; print_expr e1; print_string ")";
      )
    | (_, INT(n)) -> print_string (Big_int.string_of_big_int n)
    | (_, BOOL(b)) -> print_string (string_of_bool b)
    | (_, BLOCK(e)) -> (
	print_string "("; print_expr e; print_string ")")
    | (_, SEQ(e1, e2)) -> (
	print_string "( "; print_expr e1; print_string " ; ";
	print_expr e2; print_string " )")
    | _ -> print_string "<OTHER>"
	
and print_exprenv e env = 
  (
    print_string "[";
    print_env env;
    print_string " | ";
    print_expr e;
    print_string "]";
    print_newline ()
  )

and eval e env = ((*print_exprenv e env; *)
  let v = match e with
      (* classical lambda calculus *)
    | (p, VAR(s)) -> (
	let v = lookup s env in
	  match v with
	    | Fix(s', e', env') -> eval (p, e') (push env' s' v)
	    | _ -> v
      )
    | (p, APP(e1, e2)) -> (
	let v1 = eval e1 env in (
	    match v1 with
	      | Fun(s, e', env') ->
		  let v2 = eval e2 env in
		    eval (fst e1, e') (push env' s v2)
	      | _ -> raise Type_error
	  )
      )
    | (_, FUN(s, e)) -> Fun(s, snd e, env)

    (* let statements *)
    | (_, LET(s, e1, e2)) ->
	let v1 = eval e1 env in
	  eval e2 (push env s v1)
    | (_, LETREC(s, e1, e2)) ->
	eval e2 (push env s (Fix(s, snd e1, env)))

    (* unit expression *)
    | (_, UNIT) -> Unit

    (* integer related stuff *)
    | (_, INT(n)) -> Int(n)
    | (_, UMINUS(e)) -> (
	match eval e env with
	  | Int(n) -> Int(Big_int.minus_big_int n)
	  | _ -> raise Type_error
      )
    | (_, PLUS(e1, e2)) -> eval_binary_int e1 e2 env Big_int.add_big_int
    | (_, MINUS(e1, e2)) -> eval_binary_int e1 e2 env Big_int.sub_big_int
    | (_, MULT(e1, e2)) -> eval_binary_int e1 e2 env Big_int.mult_big_int
    | (_, DIV(e1, e2)) -> eval_binary_int e1 e2 env Big_int.div_big_int
    | (_, MOD(e1, e2)) -> eval_binary_int e1 e2 env Big_int.mod_big_int
    | (_, GREATER(e1, e2)) -> eval_binary_intbool e1 e2 env f_greater
    | (_, LESS(e1, e2)) -> eval_binary_intbool e1 e2 env f_less
    | (_, LEQ(e1, e2)) -> eval_binary_intbool e1 e2 env f_leq
    | (_, GEQ(e1, e2)) -> eval_binary_intbool e1 e2 env f_geq

    (* boolean related stuff *)
    | (_, BOOL(b)) -> Bool(b)
    | (_, NOT(e)) -> (
	match eval e env with
	  | Bool(b) -> Bool(not b)
	  | _ -> raise Type_error
      )
    | (_, AND(e1, e2)) -> eval_binary_bool e1 e2 env f_and
    | (_, OR(e1, e2)) -> eval_binary_bool e1 e2 env f_or
    | (_, XOR(e1, e2)) -> eval_binary_bool e1 e2 env f_xor

    | (_, ITE(e1, e2, e3)) -> let v1 = eval e1 env in (
	match v1 with
	  | Bool(b) -> (
	      if b then
		eval e2 env
	      else
		eval e3 env
	    )
	  | _ -> raise Type_error
      )
							
    (* equality *) 
    | (_, EQUAL(e1, e2)) ->
	let v1 = eval e1 env in
	let v2 = eval e2 env in
	  Bool(match (v1, v2) with
		 | (Bool(b1), Bool(b2)) -> (b1 = b2)
		 | (Int(n1), Int(n2)) -> (Big_int.eq_big_int n1 n2)
		 | (Unit, Unit) -> true
		 | _ -> raise Type_error)

    (* pointers *)
    | (_, REF(e)) ->
	let v = eval e env in
	let l = push_mem v in
	  Location(l)
    | (_, BANG(e)) ->
	let v = eval e env in (
	    match v with
	      | Location(l) -> search_mem l
	      | _ -> raise Type_error
	  )
    | (_, ASSIGN(e1, e2)) ->
	let v1 = eval e1 env in
	let v2 = eval e2 env in (
	    match v1 with
	      | Location(l) -> (update_mem l v2; Unit)
	      | _ -> raise Type_error
	  )

    (* sequences, blocks and type *)
    | (_, SEQ(e1, e2)) -> 
	let v1 = eval e1 env in (
	    match v1 with
	      | Unit -> eval e2 env
	      | _ -> raise Type_error
	  )
    | (_, BLOCK(e)) ->
	eval e env
    | (_, TYPE(e)) ->
	eval e env in
    ((*print_string "Done "; print_exprenv e env; *)v))

let typing expr =
  (* replace this function with a function that will output the
     types/type schemas of all expressions within a TYPE block *)
  "you should replace this output by your own"

let main () =
  try
    let cin = 
      if Array.length Sys.argv > 1
      then open_in Sys.argv.(1)
      else stdin
    in
    let lexbuf = Lexing.from_channel cin in
    let ast = Parser.input Lexer.token lexbuf in (
	print_expr ast;
	print_string "\nTYPES\n";
	print_string (typing ast);
	print_endline "\n";
	print_string "EVALUATION";
	print_endline "";
        print_value (eval ast []);
        print_endline ""
      )
  with
    | Eof_in_comment -> ( print_string "Comment without end"; exit 0 )

let _ = Printexc.print main ()
